/*
 * toon_OTGW_io.c
 * 
 * Code for accessing and controlling the OpenTherm gateway by Prodrive, for 
 * Eneco's toon thermostat. 
 *
 * The term OTGW was coined by Schelte Bron, from the OpenThermGateWay project, 
 * not Prodrive, Eneco, or Quby. 
 *
 * 20150725, MarcelR!
 *
 */

#define DEBUG                   1

#include <stdio.h>
#include <strings.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/select.h>
#include <fcntl.h>
#include <termios.h>
#include <sys/ioctl.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "toon_OTGW.h"
#include "toon_crc.h"

/* defines for read_from_port */

#define N_TRIES                10 /* # tries before port read is abandoned */
#define SERIAL_WAIT_SEC         0
#define SERIAL_WAIT_USEC    10000 /* 0.01 s */
#define BOOTSTRAP_MAX_TRIES     3

#define OT_CAPABILITY_RETRIES   5 /* 5 in production code */
#define MAX_OT_TEST_RETRIES     5

/* defines for room temp */

#define ROOM_TEMP_FILE_1 "/sys/devices/platform/imx-i2c.0/i2c-0/0-0048/device0/value"
#define ROOM_TEMP_FILE_2 "/sys/devices/platform/imx-i2c.0/i2c-0/0-0049/device1/value"
#define ROOM_TEMP_FILE      ROOM_TEMP_FILE_2
#define ROOM_TEMP_OFFSET    -6.0 /* default offset in Quby code, seems OK */
#define ROOM_TEMP_BUF_SIZE  16

/* OT ID definitions (as supported by toon)  */

#define ID_STATUS_MASTER_SLAVE     0
#define ID_BOILER_CTRL_SETPOINT    1
#define ID_CONFIG_MASTER           2
#define ID_CONFIG_SLAVE            3
#define ID_FAULT_FLAGS             5
#define ID_REMOTE_PAR_FLAGS        6
#define ID_REMOTE_OVERRIDE         9
#define ID_MAX_MODULATION_LVL     14 /* not mode-set originally */
#define ID_MAX_BOILER_POWER       15
#define ID_ROOM_TEMP_SETPOINT     16

#define ID_REL_MODULATION_LVL     17
#define ID_BOILER_WATER_PRESS     18
#define ID_ROOM_TEMP              24
#define ID_BOILER_WATER_TEMP      25
#define ID_DHW_TEMP               26
#define ID_OUTSIDE_TEMP           27
#define ID_RETURN_WATER_TEMP      28
#define ID_SOLAR_STORAGE_TEMP     29 /* not mode-set, originally */
#define ID_BOILER_FAN_SPEED       35
#define ID_DHW_TEMP_BOUNDARIES    48
     
#define ID_MAX_DHW_SETPOINT       56
#define ID_MAX_CH_WATER_SETPOINT  57
#define ID_OEM_CODE_VENTILATION   73
#define ID_FAILED_BURNER_STARTS  113      
#define ID_LOW_FLAME_SIGNALS     114     
#define ID_OEM_DIAG_CODE         115     
#define ID_BURNER_STARTS         116
#define ID_CH_PUMP_STARTS        117
#define ID_DHW_PUMP_STARTS       118
#define ID_DHW_BURNER_STARTS     119

#define ID_BURNER_HOURS          120
#define ID_CH_PUMP_HOURS         121
#define ID_DHW_PUMP_HOURS        122
#define ID_DHW_BURNER_HOURS      123
#define ID_SLAVE_NUMBER_TYPE     127

/* OT ID types */

#define ID_TYPE_UNKNOWN            0
#define ID_TYPE_FLAGS              1
#define ID_TYPE_FLOAT              2
#define ID_TYPE_SIGNED_CHAR        3
#define ID_TYPE_UNSIGNED_CHAR      4
#define ID_TYPE_FLAGS_UCHAR        5
#define ID_TYPE_UNSIGNED_SHORT     6

/* ID flags */

/* ID0 */

#define FLAG_FAULT                 0x0001
#define FLAG_CH_ACTIVE             0x0002
#define FLAG_DHW_ACTIVE            0x0004
#define FLAG_FLAME_ON              0x0008
#define FLAG_COOLING_ACTIVE        0x0010
#define FLAG_CH2_ACTIVE            0x0020
#define FLAG_DIAG_EVENT            0x0040
#define FLAG_RESERVED_0_7          0x0080

#define FLAG_CH_ENABLED            0x0100
#define FLAG_DHW_ENABLED           0x0200
#define FLAG_COOLING_ENABLED       0x0400
#define FLAG_OTC_ACTIVE            0x0800
#define FLAG_CH2_ENABLED           0x1000
#define FLAG_RESERVED_0_13         0x2000
#define FLAG_RESERVED_0_14         0x4000
#define FLAG_RESERVED_0_15         0x8000

/* ID3 */

#define FLAG_HAVE_DHW              0x0100
#define FLAG_CTRL_BANG_BANG        0x0200
#define FLAG_HAVE_COOLING          0x0400
#define FLAG_HAVE_DHW_BUFFER       0x0800
#define FLAG_NO_PUMP_CTRL          0x1000
#define FLAG_HAVE_CH2              0x2000

/* ID5 */

#define FLAG_SERVICE_REQUEST       0x0100
#define FLAG_LOCKOUT_RESET         0x0200
#define FLAG_LOW_WATER_PRESS       0x0400
#define FLAG_FLAME_FAULT           0x0800
#define FLAG_AIR_PRESS_FAULT       0x1000
#define FLAG_WATER_OVER_TEMP       0x2000

/* ID6 */

#define FLAG_DHW_SETPOINT_RW       0x0001
#define FLAG_CH_SETPOINT_RW        0x0002

#define FLAG_DHW_SETPOINT_ENABLED  0x0100
#define FLAG_CH_SETPOINT_ENABLED   0x0200

/* ID100 */

#define FLAG_ROOM_T_MAN_OVERRIDE   0x0001
#define FLAG_ROOM_T_PROG_OVERRIDE  0x0002



typedef struct about{
  int FW;
  int SN[4];
  int PN[3];
  char SN_str[16];
  char PN_str[16];
} about;



typedef struct status{
  int has_rebooted;
  int OT_capable;
} status;



typedef struct sample{
  float x;
  float t;
} sample;



typedef struct float_circ_buf{
  int index;
  int size;
  sample* data[]; 
} float_circ_buf;

/* linked-list member for OT ID description */

typedef struct ot_id{
  int            id;
  int            mode_flag;
  int            id_type;
  unsigned short val;
  double         t;
  struct ot_id*  next;
} ot_id;



/* serial communication with toon */

int            open_serial_port( char* );
void           init_serial_port( int, struct termios*, struct termios* );
void           close_serial_port( int, struct termios* );
int            write_Js( int );
int            read_Js( int );
unsigned char* cmd_io( int, unsigned char, unsigned char, unsigned char* );
about*         get_OTGW_info( int );
status*        get_OTGW_status( int );
int            confirm_OTGW_status( int ); 
int            handle_ack( unsigned char* );
ot_id*         set_master_modes( int );
int            enable_OT_circuitry( int );
int            set_DHW_setpoint( int );
int            poll_IDs( ot_id*, int );
int            poll_ID( ot_id*, int, int );
int            set_bit( int, int, int, unsigned char ); 

/* utility functions */
int            check_root( void );
void           dbgprintf( int, unsigned char*, char );
int            check_OTGW_packet( unsigned char* );
short          float_to_OT_float( float );
float          OT_float_to_float( short );
unsigned char  short_high_byte( short );
unsigned char  short_low_byte( short );
unsigned char* short_to_uchar_array( short );
void*          convert_OT_data( int, unsigned char, unsigned char );
unsigned char* read_n_bytes_from_port( int, int );
unsigned char* conditional_read( int );
int            read_from_port( int, unsigned char*, int );
float          get_room_temp( void );


static inline double get_timestamp( struct timeval* );
void set_normalized_timespec( struct timespec *, time_t, long long );
struct timespec timespec_sub( struct timespec, struct timespec );
struct timespec timespec_add( struct timespec, struct timespec );

#define NSEC_PER_SEC      1000000000L
#define ACQ_INTERVAL_SEC  1
#define ACQ_INTERVAL_NSEC 0

/* main code */

int main( void )
{
  int     port;
  struct termios oldtio, newtio;

  about*  about;
  status* status;
  int*    master_modes;
  ot_id*  id;
  int     ack;
  float   room_temp;
  float_circ_buf* room_T;
  int     ot_capable;
  /* control loop timing */

  struct timespec *delta;
  struct timespec acq_clock;

  struct timeval *t1;

  /* need double precision floating point for sub-millisecond timing accuracy */
  double t   = 0;
  double t_o = 0;
  double dt;
  int i;
  FILE* fp;

  /*
  short test;
  float ftest;

  float* ftmp_p;
  short* stmp_p;
  unsigned short* ustmp_p;

  struct timespec wait;


  wait.tv_sec= 0;
  wait.tv_nsec = 50000000;

  test  = float_to_OT_float( 105.9105 );
  ftest = OT_float_to_float( 0xffff );
  printf("test   = 0x%04x\n", test & 0xffff );
  printf("ftest  = %6.3f\n", ftest );

  ftmp_p = convert_OT_data( 1, 0x17, 0x04 );
  ftest = *ftmp_p;
  printf("ftest  = %6.3f, ftest = 0x%016x, ftest2 = %6.3f, ftest2 = %08x\n", 
	 ftest, &ftest, *ftmp_p, ftmp_p );
  */

  /* check if we are root, if not, exit gracefully */

  check_root();

  /* initialize serial port */

  port = open_serial_port( OTGW_PORT );
  init_serial_port( port, &oldtio, &newtio );

  ot_capable = 0;
  i = 0;

  /* start of bootstrap of OTGW, this call MUST come first */

  while (!ot_capable & (i < MAX_OT_TEST_RETRIES) )
    {
      about = get_OTGW_info( port );

      printf( "FW: %d\n", about->FW );
      printf( "PN: %s\n", about->PN_str );
      printf( "SN: %s\n", about->SN_str );
      
      /* set control modes for OT IDs */
      
      id = set_master_modes( port );
      ack = enable_OT_circuitry( port );
      printf("enable_OT_circuitry: ack = %d\n", ack );
      printf("set_master_modes   : id* = 0x%x\n", id );
      
      /* Set DHW temperature to a safe level */ 
      
      ack = set_DHW_setpoint( port );
      printf("set_DHW_setpoint: ack = %d\n", ack );
      
      /* get OTGW status */
      
      status = get_OTGW_status( port );
      
      printf("OTGW has_rebooted: %d\n", status->has_rebooted );
      printf("Boiler OT capable: %d\n", status->OT_capable );

      ot_capable = status->OT_capable;
      i++;
    }


  room_temp = get_room_temp();
  printf("room temperature : %8.3f deg. C\n", room_temp );

  /* 
   * Continue with actual boiler control. This should be an infinite loop, 
   * linux machine uptimes exceeding several years are no rare exceptions ;-)
   *
   * So: everything should be thoroughly tested for memory leakage. 
   *
   * Every byte leaked is one too many!! 
   */

  room_T        = malloc( sizeof( struct float_circ_buf ) );
  room_T->size  = ROOM_TEMP_BUF_SIZE;
  room_T->index = 0;
  room_T->data[0]  = malloc( ROOM_TEMP_BUF_SIZE * sizeof( sample ) );

  /* initialize the acquisition and control clock */

  t1    = malloc(sizeof(struct timeval));
  delta = malloc(sizeof(struct timespec));

  clock_gettime( CLOCK_REALTIME, &acq_clock );
  t_o = (double)(acq_clock.tv_sec) + 1e-9 * (double)(acq_clock.tv_nsec);

  /* set sampling interval */

  delta->tv_sec  = ACQ_INTERVAL_SEC;
  delta->tv_nsec = ACQ_INTERVAL_NSEC;

  /* T H E   M A I N   L O O P */

  for ( i = 0; i < 2; i++ )
    {      
      /* update time request and wait until next sample is due */
      
      acq_clock = timespec_add( acq_clock, *delta );
      clock_nanosleep( CLOCK_REALTIME, TIMER_ABSTIME, &acq_clock, NULL );
      
      /* time-critical stuff goes here, control measurements only, please */ 
      
      /* calculate timestamp and actual sampling interval */
      
      t   = get_timestamp( t1 );
      dt  = t - t_o;
      t_o = t;
      
      /* control actions go here */
      
      printf("timestamp: %11.6f s, dt = %8.6f\n", t, dt );     
    }

  if ( status->OT_capable )
    {
      /* OT boiler control */

      poll_IDs( id, port );

      /* fill default controller values */

      while( 1 == 0 )
	{
	  room_T->data[room_T->index]->x = get_room_temp();

	  if (room_T->index++ > ROOM_TEMP_BUF_SIZE + 1 )
	    room_T->index = 0;
	}
    }
  else
    {
      /* bang-bang boiler control */
    }


  /* reset serial port and close */

  close_serial_port( port, &oldtio );

  /* free everything to keep valgrind happy */

  free( delta );
  free( t1 );
  free( room_T->data[0] );
  free( room_T );
  free( status );
  free( about );

  printf("toon_OTGW_io: clean exit\n");
  return 0;
} /* end of main code */

/*

float lowpass( sample_t x, float cutoff )
{

  float RC = 1.0/(CUTOFF*2*3.14);
  float dt = 1.0/SAMPLE_RATE;
  float alpha = dt/(RC+dt);
  float filteredArray[numSamples];
  filteredArray[0] = data.recordedSamples[0];
  for(i=1; i<numSamples; i++){
    filteredArray[i] = filteredArray[i-1] + (alpha*(data.recordedSamples[i] - filteredArray[i-1]));
  }
  data.recordedSamples = filteredArray;
}

*/

int set_bit( int id, int port, int high_byte, unsigned char bit )
{

  return 0;
}



int poll_ID( ot_id* id, int port, int n ) 
{
  unsigned char   arg[4];
  unsigned char*  buf;
  int             ret;
  struct timeval* t;

  t = malloc( sizeof( struct timeval ));
  while ( id != NULL )
    {
      if (id->id == n)
	{
	  arg[0]  = 0;
	  arg[1]  = (unsigned char)(id->id);
	  arg[2]  = 0;
	  arg[3]  = 0;
	  
	  buf     = cmd_io( port, (CMD_M | OUTMASK), CMD_CAT, &arg[0] );
	  
	  id->t   = get_timestamp( t );
	  id->val = ((unsigned short)(buf[4]) << 8) | 
	    (unsigned short)(buf[5] & 0xff); 
	  printf( "id = %3d, t = %16.6f, val = %04x\n", id->id, id->t, 
		  id->val );
	  ret = 0;
	  break; break;
	}
      id      = id->next;
      ret = -1;
    }
  free( t );
  return ret;
}



int poll_IDs( ot_id* id, int port ) 
{
  unsigned char   arg[4];
  unsigned char*  buf;
  int             cnt;
  int i;
  FILE* fp;
  ot_id* id_old;
  struct timeval* t;
  fp = fopen( "/tmp/ID_log.txt", "w+");
  id_old = id;
  t = malloc( sizeof( struct timeval ));

  for ( i = 0; i < 1440; i++ )
    {
      id = id_old;
      cnt = 0;
      while ( id != NULL )
	{
	  arg[0]  = 0;
	  arg[1]  = (unsigned char)(id->id);
	  arg[2]  = 0;
	  arg[3]  = 0;
	  
	  buf     = cmd_io( port, (CMD_M | OUTMASK), CMD_CAT, &arg[0] );
	  
	  id->t   = get_timestamp( t );
	  id->val = ((unsigned short)(buf[4]) << 8) | 
	    (unsigned short)(buf[5] & 0xff); 
	  printf( "id = %3d, t = %16.6f, val = %04x\n", id->id, id->t, id->val );
	  fprintf( fp, "%3d, %16.6f, %04x\n", id->id, id->t, id->val );
    
	  id      = id->next;
	  cnt++;
	  usleep( 60000 );
	}
      usleep(1320000);
      confirm_OTGW_status( port );
    }
  free( t );
  fclose( fp );

  return cnt;
}



status* get_OTGW_status( port ) 
{
  unsigned char arg[4];
  unsigned char* buf;
  int n;
  status* stat;

  stat = malloc( sizeof( struct status ) );

  arg[0] = 0;
  arg[1] = 0;
  arg[2] = 0;
  arg[3] = 0;
  buf = cmd_io( port, (CMD_B | OUTMASK), CMD_J, &arg[0] );

  stat->has_rebooted = (int)( buf[5] );

  /* 
   * Check OT capability of the boiler. 
   */

  n = 0;
  stat->OT_capable = 0;

  do
    {
      usleep( 1000000 ); /* 1 s */
      buf = cmd_io( port, (CMD_B | OUTMASK), CMD_P, &arg[0] );
      stat->OT_capable = (int)( buf[5] );
      n++;
    }  while ( (n < OT_CAPABILITY_RETRIES ) & (stat->OT_capable == 0) );

  return stat;
}



int confirm_OTGW_status( port ) 
{
  unsigned char arg[4];
  unsigned char* buf;
  status* stat;

  stat = malloc( sizeof( struct status ) );

  arg[0] = 0;
  arg[1] = 0;
  arg[2] = 0;
  arg[3] = 0;
  buf = cmd_io( port, (CMD_B | OUTMASK), CMD_P, &arg[0] );
  stat->OT_capable = (int)( buf[5] );

  /* 
   * Check OT capability of the boiler. 
   */

  usleep( 100000 ); /* 0.1 s */
  buf = cmd_io( port, (CMD_B | OUTMASK), CMD_J, &arg[0] );

  stat->has_rebooted = (int)( buf[5] );
  free( stat );
  return 0;
}



float get_room_temp( void )
{
  float temp;
  FILE* fp;

  fp = fopen( ROOM_TEMP_FILE, "r" );
  if ( fp != NULL )
    {
      fscanf( fp, "%f\n", &temp );
      temp += ROOM_TEMP_OFFSET;
      fclose( fp );
    }
  else
    perror("get_room_temp:");

  return temp;
}



int enable_OT_circuitry( int port )
{
  /* switches on OT hardware */
  unsigned char arg[4];
  unsigned char* buf;
  arg[0] = 0;
  arg[1] = 0;
  arg[2] = 0;
  arg[3] = 1;
  buf = cmd_io( port, (CMD_K | OUTMASK), CMD_E, &arg[0] );

  return handle_ack( buf );
}



int set_DHW_setpoint( int port )
{

  /* 
   * Set DHW temperature to a safe level.
   * This is hard-coded to 60 deg C, which should be enough to kill any 
   * Legionella bacteria inside your DHW system, see:
   * 
   * Guidelines for Drinking-water Quality, 4th edition, WHO, 2011, pp 244-245.
   * (Not sure about regulatory demands, but they probably exist, at least in 
   * some developed countries).
   *
   * Don't lower this temperature, if you care about your thermostat users' 
   * health.
   */

  int ack;
  unsigned char arg[4];
  unsigned char* buf;
  arg[0] = 1;
  arg[1] = 56;
  arg[2] = 60;
  arg[3] = 0;
  buf = cmd_io( port, (CMD_M | OUTMASK), CMD_CB, &arg[0] );
  
  return handle_ack( buf );
}



ot_id* set_master_modes( int port )
{
  /* 
   * Sets master modes for OT functionality, stores data in a linked list of 
   * OT_ID structs 
   */
  int            i, n, cnt;
  int            ack;
  unsigned char  arg[4];
  unsigned char* buf;
  ot_id*         id;
  ot_id*         next_id; 
  ot_id*         previous_id;
  /* 
   * column 0: OT id, column 1: mode flag 
   * copied from happ_thermstat output, no idea what it really means (yet).
   */
  unsigned char modes[35][3] = {
    { ID_STATUS_MASTER_SLAVE,   ID_TYPE_FLAGS,          8 },
    { ID_BOILER_CTRL_SETPOINT,  ID_TYPE_FLOAT,          2 },
    { ID_CONFIG_MASTER,         ID_TYPE_FLAGS_UCHAR,    2 },
    { ID_CONFIG_SLAVE,          ID_TYPE_FLAGS_UCHAR,    1 },
    { ID_MAX_MODULATION_LVL,    ID_TYPE_FLOAT,          1 },
    { ID_MAX_BOILER_POWER,      ID_TYPE_UNSIGNED_CHAR,  1 },
    { ID_REL_MODULATION_LVL,    ID_TYPE_FLOAT,          1 },
    { ID_BOILER_WATER_TEMP,     ID_TYPE_FLOAT,          1 },
    { ID_FAULT_FLAGS,           ID_TYPE_FLAGS,          1 },
    { ID_REMOTE_PAR_FLAGS,      ID_TYPE_FLAGS_UCHAR,    1 },
    { ID_REMOTE_OVERRIDE,       ID_TYPE_FLOAT,          1 },
    { ID_ROOM_TEMP_SETPOINT,    ID_TYPE_FLOAT,          2 },
    { ID_BOILER_WATER_PRESS,    ID_TYPE_FLOAT,          1 },
    { ID_ROOM_TEMP,             ID_TYPE_FLOAT,          2 },
    { ID_DHW_TEMP,              ID_TYPE_FLOAT,          1 },
    { ID_OUTSIDE_TEMP,          ID_TYPE_FLOAT,          1 },
    { ID_RETURN_WATER_TEMP,     ID_TYPE_FLOAT,          1 },
    { ID_SOLAR_STORAGE_TEMP,    ID_TYPE_FLOAT,          1 },
    { ID_BOILER_FAN_SPEED,      ID_TYPE_UNKNOWN,        1 },
    { ID_DHW_TEMP_BOUNDARIES,   ID_TYPE_SIGNED_CHAR,    1 },
    { ID_MAX_DHW_SETPOINT,      ID_TYPE_FLOAT,          2 },
    { ID_MAX_CH_WATER_SETPOINT, ID_TYPE_FLOAT,          1 },
    { ID_OEM_CODE_VENTILATION,  ID_TYPE_UNKNOWN,        1 },
    { ID_FAILED_BURNER_STARTS,  ID_TYPE_UNSIGNED_SHORT, 1 },
    { ID_LOW_FLAME_SIGNALS,     ID_TYPE_UNSIGNED_SHORT, 1 },
    { ID_OEM_DIAG_CODE,         ID_TYPE_UNSIGNED_SHORT, 1 },
    { ID_BURNER_STARTS,         ID_TYPE_UNSIGNED_SHORT, 1 },
    { ID_CH_PUMP_STARTS,        ID_TYPE_UNSIGNED_SHORT, 1 },
    { ID_DHW_PUMP_STARTS,       ID_TYPE_UNSIGNED_SHORT, 1 },
    { ID_DHW_BURNER_STARTS,     ID_TYPE_UNSIGNED_SHORT, 1 },
    { ID_BURNER_HOURS,          ID_TYPE_UNSIGNED_SHORT, 1 },
    { ID_CH_PUMP_HOURS,         ID_TYPE_UNSIGNED_SHORT, 1 },
    { ID_DHW_PUMP_HOURS,        ID_TYPE_UNSIGNED_SHORT, 1 },
    { ID_DHW_BURNER_HOURS,      ID_TYPE_UNSIGNED_SHORT, 1 },
    { ID_SLAVE_NUMBER_TYPE,     ID_TYPE_UNSIGNED_CHAR,  1 },
  };

  cnt = 0;
  n = sizeof( modes )/3 ;

  for ( i = 0; i < n; i++ )
    {
      /* set master mode ID arg[1], flag arg[2] */

      arg[0] = 1;
      arg[1] = modes[i][0];
      arg[2] = modes[i][2];
      arg[3] = 0;
      buf = cmd_io( port, (CMD_K | OUTMASK), CMD_S, &arg[0] );
      ack = handle_ack( buf );
      printf("Mastermode ID %3d, flag %d: ack = %d\n", arg[1], arg[2], ack );

      if ( !ack )
	{
	  next_id            = malloc( sizeof( ot_id ) );
	  next_id->id        = (int)modes[i][0];
	  next_id->id_type   = (int)modes[i][1];
	  next_id->mode_flag = (int)modes[i][2];
	  next_id->next      = NULL;

	  if ( cnt == 0 )
	    {
	      id = next_id;
	    }
	  else
	    {
	      previous_id->next = next_id;
	    }
	  previous_id = next_id;
	  cnt++;
	}
    }
  return id;
}



int handle_ack( unsigned char* buf )
{
  /* check crc */
  if  ( buf[6] != crc8( CRC8_SEED_TOON, &buf[0], 6 ) )
    return -1;
  /* check contents */
  if ( buf[0] != CMD_B )
    return -2;
  if ( buf[1] != CMD_A )
    return -3;
  return 0;
}



about* get_OTGW_info( port )
{
  unsigned char arg[4];
  unsigned char* buffer_in;
  about* about;

  /* start OTGW bootstrap, by getting FW version  ... */ 
  
  about = malloc( sizeof( struct about ) );
  about->FW = get_fw_version ( port );

  /* .. and then do the rest of the bootstrap */

  /* disable OT */

  arg[0] = 0;
  arg[1] = 0;
  arg[2] = 0;
  arg[3] = 0;
  buffer_in = cmd_io( port, (CMD_K | OUTMASK), CMD_E, &arg[0] );
  printf("Disable OT: ack = %d\n", handle_ack( buffer_in ) );

  /* get OTGW part number */

  arg[0] = 0;
  arg[1] = 0;
  arg[2] = 0;
  arg[3] = 0;
  buffer_in = cmd_io( port, (CMD_B | OUTMASK), CMD_Y, &arg[0] );
 
  about->PN[0] = 1000 * (int)( buffer_in[4] >> 4 ) +
    100 *(int)( buffer_in[4] & 0x0f ) +
    10 * (int)( buffer_in[5] >> 4 ) +
    (int)( buffer_in[5] & 0x0f );

  arg[3] = 1;
  buffer_in = cmd_io( port, (CMD_B | OUTMASK), CMD_Y, &arg[0] );

  about->PN[1] = 1000 * (int)( buffer_in[2] >> 4 ) +
    100 *(int)( buffer_in[2] & 0x0f ) +
    10 * (int)( buffer_in[3] >> 4 ) +
    (int)( buffer_in[3] & 0x0f );
  about->PN[2] = 1000 * (int)( buffer_in[4] >> 4 ) +
    100 *(int)( buffer_in[4] & 0x0f ) +
    10 * (int)( buffer_in[5] >> 4 ) +
    (int)( buffer_in[5] & 0x0f );

  sprintf( about->PN_str, "%04d-%04d-%04d", about->PN[0], about->PN[1], 
	   about->PN[2] );

  /* get OTGW serial number */

  arg[3] = 2;
  buffer_in = cmd_io( port, (CMD_B | OUTMASK), CMD_Y, &arg[0] );

  about->SN[0] = 1000 * (int)( buffer_in[4] >> 4 ) +
    100 *(int)( buffer_in[4] & 0x0f ) +
    10 * (int)( buffer_in[5] >> 4 ) +
    (int)( buffer_in[5] & 0x0f );

  arg[3] = 3;
  buffer_in = cmd_io( port, (CMD_B | OUTMASK), CMD_Y, &arg[0] );

  about->SN[1] = 10 * (int)( buffer_in[2] >> 4 ) +
    (int)( buffer_in[2] & 0x0f );
  about->SN[2] = 100 * (int)( buffer_in[3] >> 4 ) +
    10*(int)( buffer_in[3] & 0x0f ) +
    (int)( buffer_in[4] >> 4 );
  about->SN[3] = 100 * (int)( buffer_in[4] & 0x0f ) +
    10*(int)( buffer_in[5] >> 4 ) +
    (int)( buffer_in[5] & 0x0f );
  sprintf( about->SN_str, "%d-%d-%03d-%03d", about->SN[0], 
	   about->SN[1], about->SN[2], about->SN[3] );

  return about;
}



unsigned char* cmd_io( int port, unsigned char cmd_type, unsigned char cmd, 
		       unsigned char* arg )
{
  int tries;

  unsigned char buffer_out[7];
  unsigned char* echo_in;
  static unsigned char* buffer_in;
  
  tries = 0;
  do {
    /* assemble cmd type, cmd and args + crc into one data packet */    

    memset( buffer_out, 0x0, 7 );
    buffer_out[0] = cmd_type | 0x80;
    buffer_out[1] = cmd;
    memcpy( &buffer_out[2], &arg[0], 4 );
    buffer_out[6] = crc8( CRC8_SEED_TOON, &buffer_out[0], 6 );

    write( port, &buffer_out, 7 ); 
    dbgprintf( DEBUG, &buffer_out[0], '>' );
    
    echo_in = conditional_read( port );
    dbgprintf( DEBUG, echo_in, '<' );
    buffer_in = conditional_read( port );
    dbgprintf( DEBUG, buffer_in, '<' );
    
    if (!check_OTGW_packet( buffer_in ))
      {
	return &buffer_in[0];
      }

    tries++;
  } while ( tries < BOOTSTRAP_MAX_TRIES );
  
}



int get_fw_version( int port )
{
  /*
   * Read FW version from OTGW. The IO protocol is different from all other
   * IO requests (don't know why), so that's why this is a separate routine. 
   */
  int tries;
  int jw, jr;

  unsigned char buffer_out[7];
  unsigned char* echo_in;
  unsigned char* buffer_in;
  
  tries = 0;
  do {
    
    jw = write_Js( port );
    /* 
     * write fw version request command to port, 
     * (B & 0x80 V 0 0 0 0 )
     * Apparently, this command goes without crc. Possibly historical reasons
     * (like no crc check done in early FW versions).
     */
    memset( buffer_out, 0x0, 7 );
    buffer_out[0] = CMD_B | 0x80;
    buffer_out[1] = CMD_V;
    write( port, &buffer_out, 6 ); 
    dbgprintf( DEBUG, &buffer_out[0], '>' );
    
    jr = read_Js( port );
    echo_in = conditional_read( port );
    dbgprintf( DEBUG, echo_in, '<' );
    buffer_in = conditional_read( port );
    dbgprintf( DEBUG, buffer_in, '<' );
    
    if (!check_OTGW_packet( buffer_in ))
      {
	return (int)buffer_in[5];
      }

    tries++;
  } while ( ( (jw < 128) | (jr < 128) ) & (tries < BOOTSTRAP_MAX_TRIES) );
  
}



unsigned char* conditional_read( int port )
{
  struct timeval timeout;
  int ret;
  int cnt;
  static unsigned char buf[7];
  fd_set rdset;
  int maxfd;

  /* read first byte at serial port */

   ret = read_from_port( port, &buf[0], 1 );

  if (ret <= 0 )
    return NULL;

  if ( (buf[0] == CMD_B) | (buf[0] == CMD_K) | ( buf[0] == CMD_M) | 
       (buf[0] == (CMD_B | OUTMASK )) | (buf[0] == (CMD_K | OUTMASK)) | 
       (buf[0] == (CMD_M | OUTMASK)) )
    {
      /* cmd-type byte read, read the actual command (5 bytes) */
      ret = read( port, &buf[1], 5 );

      /* check if this is a firmware version request cmd */
      if ( (buf[0] == (CMD_B | OUTMASK)) &
	   (buf[1] == CMD_V) )
	{
	  return &buf[0];
	}
      else
	{
	  /* read crc byte */
	  ret = read_from_port( port, &buf[6], 1 );
	  
	  return &buf[0];
	}
    }
  else
    {
      return &buf[0];
    }  
}



int read_from_port( int port, unsigned char* buf, int n )
{
  /*
   * reads n bytes from port port, and stores them in buf.
   * It is up to the programmer to make the buffer sufficiently large.
   *
   * returns the number of bytes read, or a negative number in case of error.
   */
  int ret;
  int cnt;
  struct timeval timeout;
  fd_set rdset;
  int maxfd;
  int i;
  ret = -1;
  cnt = 0;
  
  /* read bytes at serial port */
  
  while ((cnt < N_TRIES) & (ret <=0 ))
    {
      timeout.tv_sec = SERIAL_WAIT_SEC;
      timeout.tv_usec = SERIAL_WAIT_USEC;
      
      FD_ZERO( &rdset );
      FD_SET( port, &rdset );
      maxfd = port + 1;
      
      ret = select( maxfd, &rdset, NULL, NULL, &timeout );
      cnt++;
    }
  switch ( ret ) {
  case 0:
    perror("read_from_port: ");
    return -1;
    break;
  default:
    ret < 0 ? ( perror("read_from_port: ") ) : 
    ( i = read( port, buf, n ) );
    break;
  }

  if (ret < 0 )
    return -2;
  else
    return i;
}



int write_Js( int port )
/* Write 128 times 'j' to serial port */
{
  int i;
  int cnt;
  char buf = CMD_j;
  cnt = 0;
  for ( i = 0; i < 128; i++ )
    {
      cnt += write( port, &buf, 1 );
    }
  return cnt;
}



int read_Js( int port )
/* Read 128 times 'j' from serial port */
{
  int i;
  unsigned char j;
  int cnt, cnt2;
  struct timeval timeout;
  fd_set rdset;
  int maxfd;

  int ret;
  cnt = 0;
  for ( i = 0; i < 128; i++ )
    {
      ret = -1;
      cnt2 = 0;
      while ((cnt2 < 100) & (ret <=0 ))
	{
	  timeout.tv_sec = SERIAL_WAIT_SEC;
	  timeout.tv_usec = SERIAL_WAIT_USEC; /* 0.01 s */
	  
	  FD_ZERO( &rdset );
	  FD_SET( port, &rdset );
	  maxfd = port + 1;

	  ret = select( maxfd, &rdset, NULL, NULL, &timeout );
	  cnt2++;
	}
      switch ( ret ) {
      case 0:
	//printf("try # %3d, select() timed out\n", i );
	break;
      default:
	ret < 0 ? ( perror("read_Js") ) : (cnt += read( port, &j, 1 ));
	//	printf("readJs: read # %3d, %c, %02X, tries = %d\n", i, j, j, cnt2 );
	break;
      }
    }
  return cnt;
}



unsigned char* read_n_bytes_from_port( int port, int n )
{  
  static unsigned char buf[14];
  struct timeval timeout;
  fd_set rdset;
  int maxfd;
  
  int sel_ret;
  int cnt;
  int i;
  
  sel_ret = -1;
  cnt = 0;
  
  while ((sel_ret <= 0) & (cnt < 10))
    {
      timeout.tv_sec = 0;
      timeout.tv_usec = 10000; /* 0.01 s */
      
      FD_ZERO( &rdset );
      FD_SET( port, &rdset );
      maxfd = port + 1;

      sel_ret = select( maxfd, &rdset, NULL, NULL, &timeout );
      cnt++;
    }
  if ( sel_ret > 0 )
    {
      read( port, &buf, n );
      printf("read from port %d:", port );
      for (i = 0; i< n; i++)
	{
	  printf(" %02X", buf[i] );
	}
      printf("\n");
    }
  else
    {
      perror("no bytes read, timeout or error");
    }
}



int check_OTGW_packet( unsigned char* buf )
{
  /* 
   * crc8 check on a 6-byte OTGW packet. 
   */
  int echo;

  buf[0] >= OUTMASK  ? (echo = 1) : (echo = 0);
  if (buf[6] == crc8(CRC8_SEED_TOON, buf, 6)) 
    return 0; 
  else 
    return -1;
}



void dbgprintf( int debug, unsigned char* buf, char dir )
{
  int i;
  int n;
  if ( (buf[0] == (CMD_B|OUTMASK)) & (buf[1] == CMD_V) )
    n = 6;
  else
    n = 7;

  if ( debug )
    {
      (dir == '>') ? printf( "OTGW <<- " ) : printf( "OTGW ->> " );
      for (i = 0; i < n; i++ )
	{
	  (dir == '>') ? printf("%02X ", buf[i] ) : printf("%02x ", buf[i] );
	}
      printf ("\n");
    }
}



int open_serial_port( char* device )
/* 
 * open_serial_port opens serial port with device special file device,
 * e.g., /dev/ttyUSB0, and on success returns a valid file descriptor for this
 * port.
 */
{
  int fd;
  /* 
   * Open serial device for reading and writing and not as controlling tty
   * because we don't want to get killed if linenoise sends CTRL-C.
   */
  fd = open( device, O_RDWR | O_NOCTTY );
  if ( fd < 0 ) 
    { 
      perror( device ); 
      exit( EXIT_FAILURE ); 
    } else 
    {
      return fd;
    }
}



void init_serial_port( int fd, struct termios* oldtio, struct termios* newtio )
/*
 * init_serial_port initializes serial port with file descriptor fd for use 
 * with a toon thermostat.
 * The original port settings are stored in termios struct oldtio, the new
 * port configuration is stored in newtio.
 *
 * Configuration parameters have been copied from a working toon, sw version 
 * 2.9.15, connected to a Nefit EMS module, which in turn was connected to a 
 * Nefit HRC30 topline boiler.
 */

{
  tcgetattr( fd, oldtio );            /* save current serial port settings.
					 Not really needed here, the io-port 
					 for OTGW is hard-wired to 
					 /dev/ttymxc1  */

  memset( newtio, 0x0, sizeof( *newtio ) ); /* clear struct for new port settings */

  /*
   * Initialize input modes.
   */

  newtio->c_iflag = 0;

  /*
   * Initialize output modes, no idea why these are set, you don't need them
   * in raw transfers.
   *
   * NLDLY  Newline delay mask. Values are NL0 and NL1.
   * CRDLY  Carriage return delay mask. Values are CR0, CR1, CR2, or CR3.
   * TABDLY Horizontal  tab  delay  mask.  Values are TAB0, TAB1, TAB2, TAB3
   *        (or XTABS).  A value of TAB3, that is, XTABS,  expands  tabs  to
   *        spaces   (with   tab  stops  every  eight  columns).
   * BSDLY  Backspace delay mask. Values are BS0 or BS1. (Has  never  been
   *        implemented.) 
   * VTDLY  Vertical tab delay mask. Values are VT0 or VT1.
   * FFDLY  Form feed delay  mask. Values are FF0 or FF1.
   */

  newtio->c_oflag = NL0 | CR0 | TAB0 | BS0 | VT0 | FF0;  

  /*
   * Initialize control modes.
   *
   * "standard" port settings: 
   * B4800   : Set baud rate. 4800 for toon's OTGW.
   * CS8     : 8n1 (8bit,no parity,1 stopbit)
   * CLOCAL  : local connection, no modem control
   * CREAD   : enable receiving characters
   * PARENB  : Parity checking?
   */

  newtio->c_cflag = B4800 | CS8 | CLOCAL | CREAD | PARENB;

  /*
   * Initialize local modes.
   */

  newtio->c_lflag = 0;

  /* 
   * Initialize control characters 
   *
   * VTIME    : Timeout in deciseconds for non-canonical read.
   * VMIN     : Minimum number of characters for non-canonical read.
   */

  newtio->c_cc[VTIME]  = 1;        /* inter-character timer used         */
  newtio->c_cc[VMIN]   = 5;        /* blocking read until 5 char arrive. */

  /* 
   * Activate settings 
   */ 

  tcflush( fd, TCIFLUSH );
  tcsetattr( fd, TCSANOW, newtio );
}



void close_serial_port( int fd, struct termios* tio )
/*
 * close_serial_port() closes serial port with file descriptor fd and restores
 * the port settings to the ones stored in tio.
 */
{
  tcsetattr( fd, TCSANOW, tio );
  close( fd );
}



int check_root( void )
/* 
 * test if the program runs with suid root. If not, exit. 
 */
{
  if ( getuid() != 0 )
    {
      printf( "This program must be run by root (user id 0).\n" );
      printf( "Using sudo is fine too ...\n " );
      exit(-1);
    } 
  else 
    {
      return 0;
    }
}



short float_to_OT_float( float val )
{
  /*
   * helper function for transforming a (normal) float to a signed short, 
   * posing as a float, according to the OT protocol.
   * Useful for filling in real numbers into OT commands (f8.8-style). 
   * Just split the short into two separate unsigned chars and you're done.  
   */

  return (short)(val * 256.0);
}



float OT_float_to_float( short val )
{
  /* 
   * convert signed short (f8.8 float really) to float according to the 
   * OT protocol. 
   */

  return ((float) val ) / 256.0;
}



unsigned char* short_to_uchar_array( short val )
{
}



unsigned char short_high_byte( short val )
{
  return (unsigned char)( (val & 0xff00 ) >> 8 );
}



unsigned char short_low_byte( short val )
{
  return (unsigned char)( val & 0xff );
}



void* convert_OT_data( int id, unsigned char high_byte, unsigned char low_byte )
{
  static unsigned short val_us;
  static short val_s;
  static float val_f;

  switch( id ) {
    
    /* two bytes filled with switches, 16 bits total: */
    
  case 0: /* status byte: high byte: master, low byte: slave                */
  case 2: /* high byte: master configuration, low byte: master member ID    */
  case 3: /* high byte: slave configuration, low byte: slave member ID      */
  case 5: /* high byte: app-specific fault flags, low byte: OEM fault code  */
  case 6: /* high byte: remote par enable flag, low byte: remote par r/w 
	     flag */
    
    /* 
     * order of bits yet unclear, order in the array is now:
     * high byte: 0, 1, ... 7, low byte 0, 1, ... 7.
     */ 

    val_us = ((unsigned short)(high_byte) << 8) + (unsigned short)low_byte;
    return &val_us;
    break; 
 
    /* 
     * f8.8: fixed point, 
     * value = (signed high byte << 8 + unsigned low byte)/256 
     */
    
  case 1:  /* boiler control setpoint [deg C]      */
  case 14: /* maximum modulation level setting [%] */ 
  case 15: /* maximum boiler power [kW]            */
  case 16: /* room temperature setpoint [deg C]    */
  case 17: /* relative modulation level [0-100%]   */
  case 18: /* boiler water pressure [bar]          */
  case 24: /* room temperature [deg C]             */
  case 25: /* boiler water temperature [deg C]     */
  case 26: /* DHW temperature [deg C]              */
  case 27: /* Outside temperature [deg C]          */
  case 28: /* return water temperature [deg C]     */
  case 29: /* solar storage temperature [deg C]    */
  case 56: /* max DHW setpoint [deg C]             */
  case 57: /* max CH water setpoint [deg C]        */

    val_f = OT_float_to_float( ((short)(high_byte) << 8) + (short)low_byte );
    return &val_f;
    break;
    
    /* two signed characters */
    
  case 48: /* high byte: DHW temp high bound, low: DHW temp low bound */
    
    val_s = ((short)(high_byte) << 8) + (short)low_byte;
    return &val_s;
    break;
    
    /* unsigned short */
    
  case 113: /* number of unsuccessful burner starts     */
  case 114: /* number of times flame signal was too low */
  case 115: /* OEM diagnostic code                      */
  case 116: /* number of burner starts                  */
  case 117: /* number of CH pump starts                 */
  case 118: /* number of DHW pump/valve starts          */
  case 119: /* number of burner starts for DHW          */
  case 120: /* burner operation hours                   */
  case 121: /* CH pump hours                            */
  case 122: /* DHW pump/valve hours                     */
  case 123: /* DHW burner hours                         */
    
    val_us = ((unsigned short)(high_byte) << 8) + (unsigned short)low_byte;
    return &val_us;
    break;

    /* two unsigned characters */
    
  case 127: /*slave number and type */
    val_us = ((unsigned short)(high_byte) << 8) + (unsigned short)low_byte;
    return &val_us;
    break;
    
    /* parameters with unknown format */
  case 73: /* OEM-specific code for ventilation / heat recovery */
  default:
    
    /* store as unsigned short */
    val_us = ((unsigned short)(high_byte) << 8) + (unsigned short)low_byte;
    return &val_us;    
    break;
  }
}


unsigned char crc8( unsigned char crc, unsigned char* data, size_t len )
/* 
 * code grabbed from the internet (www.stackoverflow.com). 
 *
 * http://stackoverflow.com/questions/15169387/definitive-crc-for-c, retrieved
 * 20150712.
 *
 * minor changes applied:
 * The seed was changed to 0xAD (see toon_crc8.h). 
 * Identifier crc8_table changed to crc8_table_toon. 
 *
 * 8-bit CRC with polynomial x^8+x^6+x^3+x^2+1, 0x15A.
 * Chosen based on Koopman, et al. (0xAD in his notation = 0x15A >> 1):
 * http://www.ece.cmu.edu/~koopman/roses/dsn04/koopman04_crc_poly_embedded.pdf
 */
{
    unsigned char* end;

    if ( len == 0 )
      return crc;
    crc ^= 0xff;
    end = data + len;
    do {
      crc = crc8_table_toon[crc ^ *data++];
    } while ( data < end );
    return crc ^ 0xff;
}



/* 
 * timing functions (partly taken from kernel code)
 */


static inline double get_timestamp( struct timeval *t ) 
{

  /* 
   * calculate timestamp
   * struct t needs to be malloc'd prior to calling this function.
   * double precision needed for sub-millisecond timing accuracy.
   */
  
  gettimeofday( t, NULL );
  
  return (double)( t->tv_sec ) + 1e-6 * (double)( t->tv_usec );
}


/* timespec stuff to use ns time resolution */

void set_normalized_timespec( struct timespec *ts, time_t sec, long long nsec )
{
  while (nsec >= NSEC_PER_SEC)
  {
    /*
    * The following asm() prevents the compiler from
    * optimising this loop into a modulo operation. See
    * also __iter_div_u64_rem() in include/linux/time.h
    */

    asm("" : "+rm"(nsec));
    nsec -= NSEC_PER_SEC;
    ++sec;
  }

  while (nsec < 0)
  {
    asm("" : "+rm"(nsec));
    nsec += NSEC_PER_SEC;
    --sec;
  }
  ts->tv_sec = sec;
  ts->tv_nsec = nsec;
}



struct timespec timespec_sub(struct timespec lhs,struct timespec rhs)
{
  struct timespec ts_delta;
  /* subtract normalized time and return a struct timespec */

  set_normalized_timespec( &ts_delta, lhs.tv_sec - rhs.tv_sec,
			   lhs.tv_nsec - rhs.tv_nsec );
  return ts_delta;
}



struct timespec timespec_add(struct timespec lhs, struct timespec rhs)
{
  /* add normalized time and return a struct timespec */

  struct timespec ts_delta; 
  set_normalized_timespec( &ts_delta, lhs.tv_sec + rhs.tv_sec, 
			   lhs.tv_nsec + rhs.tv_nsec );
  return ts_delta;
}
